# -*- coding: utf-8 -*-
#
# Picard, the next-generation MusicBrainz tagger
#
# Copyright (C) 2008-2009, 2020-2021 Philipp Wolfer
# Copyright (C) 2012-2013 Michael Wiencek
# Copyright (C) 2018, 2020-2021 Laurent Monin
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.


from picard.config import (
    BoolOption,
    IntOption,
    TextOption,
    get_config,
)

from picard.ui.options import (
    OptionsPage,
    register_options_page,
)
from picard.ui.ui_options_ratings import Ui_RatingsOptionsPage


class RatingsOptionsPage(OptionsPage):

    NAME = 'ratings'
    TITLE = N_("Ratings")
    PARENT = 'metadata'
    SORT_ORDER = 20
    ACTIVE = True
    HELP_URL = "/config/options_ratings.html"

    options = [
        BoolOption('setting', 'enable_ratings', False),
        TextOption('setting', 'rating_user_email', 'users@musicbrainz.org'),
        BoolOption('setting', 'submit_ratings', True),
        IntOption('setting', 'rating_steps', 6),
    ]

    def __init__(self, parent=None):
        super().__init__(parent)
        self.ui = Ui_RatingsOptionsPage()
        self.ui.setupUi(self)

    def load(self):
        config = get_config()
        self.ui.enable_ratings.setChecked(config.setting['enable_ratings'])
        self.ui.rating_user_email.setText(config.setting['rating_user_email'])
        self.ui.submit_ratings.setChecked(config.setting['submit_ratings'])

    def save(self):
        config = get_config()
        config.setting['enable_ratings'] = self.ui.enable_ratings.isChecked()
        config.setting['rating_user_email'] = self.ui.rating_user_email.text()
        config.setting['submit_ratings'] = self.ui.submit_ratings.isChecked()


register_options_page(RatingsOptionsPage)



# -*- coding: utf-8 -*-
#
# Picard, the next-generation MusicBrainz tagger
#
# Copyright (C) 2006-2007 Lukáš Lalinský
# Copyright (C) 2011 Calvin Walton
# Copyright (C) 2013, 2018-2021 Laurent Monin
# Copyright (C) 2016-2017 Sambhav Kothari
# Copyright (C) 2018 Wieland Hoffmann
# Copyright (C) 2019 Philipp Wolfer
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.


"""A webbrowser extension for Picard.

It handles and displays errors in PyQt and also adds a utility function for opening Picard URLS.
"""

from sys import version_info
import webbrowser

from PyQt5 import QtWidgets

from picard import log
from picard.const import PICARD_URLS


def open(url):
    if url in PICARD_URLS:
        url = PICARD_URLS[url]
    try:
        webbrowser.open(url)
    except webbrowser.Error as e:
        QtWidgets.QMessageBox.critical(None, _("Web Browser Error"), _("Error while launching a web browser:\n\n%s") % (e,))
    except TypeError:
        if version_info.major == 3 and version_info.minor == 7 and version_info.micro == 0:
            # See https://bugs.python.org/issue31014, webbrowser.open doesn't
            # work on 3.7.0 the first time it's called. The initialization code
            # in it will be skipped after the first call, making it possibly to
            # use it, although it might not accurately identify the users
            # preferred browser.
            log.info("Working around https://bugs.python.org/issue31014 - URLs might not be opened in the correct browser")
            webbrowser.open(url)



# -*- coding: utf-8 -*-

# Automatically generated - don't edit.
# Use `python setup.py build_ui` to update it.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_ScriptEditor(object):
    def setupUi(self, ScriptEditor):
        ScriptEditor.setObjectName("ScriptEditor")
        ScriptEditor.resize(902, 729)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(ScriptEditor.sizePolicy().hasHeightForWidth())
        ScriptEditor.setSizePolicy(sizePolicy)
        self.verticalLayout_3 = QtWidgets.QVBoxLayout(ScriptEditor)
        self.verticalLayout_3.setContentsMargins(0, 0, 0, 0)
        self.verticalLayout_3.setObjectName("verticalLayout_3")
        self.layout_for_menubar = QtWidgets.QHBoxLayout()
        self.layout_for_menubar.setSpacing(0)
        self.layout_for_menubar.setObjectName("layout_for_menubar")
        self.verticalLayout_3.addLayout(self.layout_for_menubar)
        self.content_layout = QtWidgets.QVBoxLayout()
        self.content_layout.setContentsMargins(9, 0, 9, 9)
        self.content_layout.setObjectName("content_layout")
        self.splitter_between_editor_and_examples = QtWidgets.QSplitter(ScriptEditor)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.MinimumExpanding, QtWidgets.QSizePolicy.MinimumExpanding)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.splitter_between_editor_and_examples.sizePolicy().hasHeightForWidth())
        self.splitter_between_editor_and_examples.setSizePolicy(sizePolicy)
        self.splitter_between_editor_and_examples.setMinimumSize(QtCore.QSize(0, 0))
        self.splitter_between_editor_and_examples.setFrameShape(QtWidgets.QFrame.NoFrame)
        self.splitter_between_editor_and_examples.setOrientation(QtCore.Qt.Vertical)
        self.splitter_between_editor_and_examples.setObjectName("splitter_between_editor_and_examples")
        self.frame_4 = QtWidgets.QFrame(self.splitter_between_editor_and_examples)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.MinimumExpanding, QtWidgets.QSizePolicy.MinimumExpanding)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(5)
        sizePolicy.setHeightForWidth(self.frame_4.sizePolicy().hasHeightForWidth())
        self.frame_4.setSizePolicy(sizePolicy)
        self.frame_4.setMinimumSize(QtCore.QSize(0, 250))
        self.frame_4.setFrameShape(QtWidgets.QFrame.NoFrame)
        self.frame_4.setFrameShadow(QtWidgets.QFrame.Raised)
        self.frame_4.setObjectName("frame_4")
        self.verticalLayout_5 = QtWidgets.QVBoxLayout(self.frame_4)
        self.verticalLayout_5.setContentsMargins(0, 3, 0, 0)
        self.verticalLayout_5.setObjectName("verticalLayout_5")
        self.horizontalLayout_2 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_2.setContentsMargins(-1, -1, -1, 0)
        self.horizontalLayout_2.setObjectName("horizontalLayout_2")
        self.label = QtWidgets.QLabel(self.frame_4)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Maximum, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.label.sizePolicy().hasHeightForWidth())
        self.label.setSizePolicy(sizePolicy)
        self.label.setMinimumSize(QtCore.QSize(0, 0))
        self.label.setIndent(0)
        self.label.setObjectName("label")
        self.horizontalLayout_2.addWidget(self.label)
        self.preset_naming_scripts = QtWidgets.QComboBox(self.frame_4)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.MinimumExpanding, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.preset_naming_scripts.sizePolicy().hasHeightForWidth())
        self.preset_naming_scripts.setSizePolicy(sizePolicy)
        self.preset_naming_scripts.setMinimumSize(QtCore.QSize(200, 0))
        self.preset_naming_scripts.setObjectName("preset_naming_scripts")
        self.horizontalLayout_2.addWidget(self.preset_naming_scripts)
        self.verticalLayout_5.addLayout(self.horizontalLayout_2)
        self.frame = QtWidgets.QFrame(self.frame_4)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.MinimumExpanding)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.frame.sizePolicy().hasHeightForWidth())
        self.frame.setSizePolicy(sizePolicy)
        self.frame.setMinimumSize(QtCore.QSize(0, 200))
        self.frame.setFrameShape(QtWidgets.QFrame.NoFrame)
        self.frame.setFrameShadow(QtWidgets.QFrame.Raised)
        self.frame.setObjectName("frame")
        self.verticalLayout_8 = QtWidgets.QVBoxLayout(self.frame)
        self.verticalLayout_8.setContentsMargins(0, 0, 0, 0)
        self.verticalLayout_8.setObjectName("verticalLayout_8")
        self.splitter_between_editor_and_documentation = QtWidgets.QSplitter(self.frame)
        self.splitter_between_editor_and_documentation.setMinimumSize(QtCore.QSize(0, 0))
        self.splitter_between_editor_and_documentation.setFrameShape(QtWidgets.QFrame.NoFrame)
        self.splitter_between_editor_and_documentation.setOrientation(QtCore.Qt.Horizontal)
        self.splitter_between_editor_and_documentation.setOpaqueResize(True)
        self.splitter_between_editor_and_documentation.setObjectName("splitter_between_editor_and_documentation")
        self.frame_5 = QtWidgets.QFrame(self.splitter_between_editor_and_documentation)
        self.frame_5.setFrameShape(QtWidgets.QFrame.NoFrame)
        self.frame_5.setFrameShadow(QtWidgets.QFrame.Raised)
        self.frame_5.setObjectName("frame_5")
        self.verticalLayout_2 = QtWidgets.QVBoxLayout(self.frame_5)
        self.verticalLayout_2.setContentsMargins(0, 0, 0, 0)
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setContentsMargins(-1, 0, -1, -1)
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.label_2 = QtWidgets.QLabel(self.frame_5)
        self.label_2.setObjectName("label_2")
        self.horizontalLayout.addWidget(self.label_2)
        self.script_title = QtWidgets.QLineEdit(self.frame_5)
        self.script_title.setObjectName("script_title")
        self.horizontalLayout.addWidget(self.script_title)
        self.verticalLayout_2.addLayout(self.horizontalLayout)
        self.file_naming_format = ScriptTextEdit(self.frame_5)
        self.file_naming_format.setEnabled(False)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.MinimumExpanding)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(5)
        sizePolicy.setHeightForWidth(self.file_naming_format.sizePolicy().hasHeightForWidth())
        self.file_naming_format.setSizePolicy(sizePolicy)
        self.file_naming_format.setMinimumSize(QtCore.QSize(0, 50))
        self.file_naming_format.viewport().setProperty("cursor", QtGui.QCursor(QtCore.Qt.IBeamCursor))
        self.file_naming_format.setTabChangesFocus(False)
        self.file_naming_format.setLineWrapMode(QtWidgets.QTextEdit.NoWrap)
        self.file_naming_format.setTabStopWidth(20)
        self.file_naming_format.setAcceptRichText(False)
        self.file_naming_format.setTextInteractionFlags(QtCore.Qt.TextEditorInteraction)
        self.file_naming_format.setObjectName("file_naming_format")
        self.verticalLayout_2.addWidget(self.file_naming_format)
        self.documentation_frame = QtWidgets.QFrame(self.splitter_between_editor_and_documentation)
        self.documentation_frame.setMinimumSize(QtCore.QSize(100, 0))
        self.documentation_frame.setFrameShape(QtWidgets.QFrame.NoFrame)
        self.documentation_frame.setFrameShadow(QtWidgets.QFrame.Raised)
        self.documentation_frame.setObjectName("documentation_frame")
        self.documentation_frame_layout = QtWidgets.QVBoxLayout(self.documentation_frame)
        self.documentation_frame_layout.setContentsMargins(0, 0, 0, 0)
        self.documentation_frame_layout.setObjectName("documentation_frame_layout")
        self.verticalLayout_8.addWidget(self.splitter_between_editor_and_documentation)
        self.verticalLayout_5.addWidget(self.frame)
        self.renaming_error = QtWidgets.QLabel(self.frame_4)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.MinimumExpanding, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.renaming_error.sizePolicy().hasHeightForWidth())
        self.renaming_error.setSizePolicy(sizePolicy)
        self.renaming_error.setText("")
        self.renaming_error.setAlignment(QtCore.Qt.AlignCenter)
        self.renaming_error.setObjectName("renaming_error")
        self.verticalLayout_5.addWidget(self.renaming_error)
        self.groupBox = QtWidgets.QGroupBox(self.splitter_between_editor_and_examples)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.MinimumExpanding, QtWidgets.QSizePolicy.MinimumExpanding)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(1)
        sizePolicy.setHeightForWidth(self.groupBox.sizePolicy().hasHeightForWidth())
        self.groupBox.setSizePolicy(sizePolicy)
        self.groupBox.setMinimumSize(QtCore.QSize(300, 150))
        self.groupBox.setMaximumSize(QtCore.QSize(16777215, 400))
        self.groupBox.setBaseSize(QtCore.QSize(0, 150))
        self.groupBox.setObjectName("groupBox")
        self.verticalLayout = QtWidgets.QVBoxLayout(self.groupBox)
        self.verticalLayout.setContentsMargins(3, 3, 3, 3)
        self.verticalLayout.setObjectName("verticalLayout")
        self.splitter_between_before_and_after = QtWidgets.QSplitter(self.groupBox)
        self.splitter_between_before_and_after.setOrientation(QtCore.Qt.Horizontal)
        self.splitter_between_before_and_after.setOpaqueResize(True)
        self.splitter_between_before_and_after.setObjectName("splitter_between_before_and_after")
        self.frame_2 = QtWidgets.QFrame(self.splitter_between_before_and_after)
        self.frame_2.setFrameShape(QtWidgets.QFrame.NoFrame)
        self.frame_2.setFrameShadow(QtWidgets.QFrame.Raised)
        self.frame_2.setObjectName("frame_2")
        self.verticalLayout_4 = QtWidgets.QVBoxLayout(self.frame_2)
        self.verticalLayout_4.setContentsMargins(0, 0, 0, 0)
        self.verticalLayout_4.setSpacing(3)
        self.verticalLayout_4.setObjectName("verticalLayout_4")
        self.example_filename_before_label = QtWidgets.QLabel(self.frame_2)
        self.example_filename_before_label.setObjectName("example_filename_before_label")
        self.verticalLayout_4.addWidget(self.example_filename_before_label)
        self.example_filename_before = QtWidgets.QListWidget(self.frame_2)
        self.example_filename_before.setEditTriggers(QtWidgets.QAbstractItemView.CurrentChanged|QtWidgets.QAbstractItemView.DoubleClicked|QtWidgets.QAbstractItemView.EditKeyPressed|QtWidgets.QAbstractItemView.SelectedClicked)
        self.example_filename_before.setSelectionBehavior(QtWidgets.QAbstractItemView.SelectRows)
        self.example_filename_before.setObjectName("example_filename_before")
        self.verticalLayout_4.addWidget(self.example_filename_before)
        self.frame_3 = QtWidgets.QFrame(self.splitter_between_before_and_after)
        self.frame_3.setFrameShape(QtWidgets.QFrame.NoFrame)
        self.frame_3.setFrameShadow(QtWidgets.QFrame.Raised)
        self.frame_3.setObjectName("frame_3")
        self.verticalLayout_6 = QtWidgets.QVBoxLayout(self.frame_3)
        self.verticalLayout_6.setContentsMargins(0, 0, 0, 0)
        self.verticalLayout_6.setSpacing(3)
        self.verticalLayout_6.setObjectName("verticalLayout_6")
        self.example_filename_after_label = QtWidgets.QLabel(self.frame_3)
        self.example_filename_after_label.setObjectName("example_filename_after_label")
        self.verticalLayout_6.addWidget(self.example_filename_after_label)
        self.example_filename_after = QtWidgets.QListWidget(self.frame_3)
        self.example_filename_after.setEditTriggers(QtWidgets.QAbstractItemView.CurrentChanged|QtWidgets.QAbstractItemView.DoubleClicked|QtWidgets.QAbstractItemView.EditKeyPressed|QtWidgets.QAbstractItemView.SelectedClicked)
        self.example_filename_after.setSelectionBehavior(QtWidgets.QAbstractItemView.SelectRows)
        self.example_filename_after.setObjectName("example_filename_after")
        self.verticalLayout_6.addWidget(self.example_filename_after)
        self.verticalLayout.addWidget(self.splitter_between_before_and_after)
        self.content_layout.addWidget(self.splitter_between_editor_and_examples)
        self.buttonbox = QtWidgets.QDialogButtonBox(ScriptEditor)
        self.buttonbox.setObjectName("buttonbox")
        self.content_layout.addWidget(self.buttonbox)
        self.verticalLayout_3.addLayout(self.content_layout)

        self.retranslateUi(ScriptEditor)
        QtCore.QMetaObject.connectSlotsByName(ScriptEditor)

    def retranslateUi(self, ScriptEditor):
        _translate = QtCore.QCoreApplication.translate
        self.label.setText(_("Selected file naming script:"))
        self.preset_naming_scripts.setToolTip(_("Select the file naming script to load into the editor"))
        self.label_2.setText(_("Title:"))
        self.groupBox.setTitle(_("Files will be named like this:"))
        self.example_filename_before_label.setText(_("Before"))
        self.example_filename_after_label.setText(_("After"))
from picard.ui.widgets.scripttextedit import ScriptTextEdit



# -*- coding: utf-8 -*-
#
# Picard, the next-generation MusicBrainz tagger
#
# Copyright (C) 2017 Antonio Larrosa
# Copyright (C) 2018-2021 Laurent Monin
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.


from collections.abc import MutableMapping


class LRUCache(MutableMapping):
    """
    Helper class to cache items using a Least Recently Used policy.

    It's originally used to cache generated pixmaps in the CoverArtBox object
    but it's generic enough to be used for other purposes if necessary.
    The cache will never hold more than max_size items and the item least
    recently used will be discarded.

    >>> cache = LRUCache(3)
    >>> cache['item1'] = 'some value'
    >>> cache['item2'] = 'some other value'
    >>> cache['item3'] = 'yet another value'
    >>> cache['item1']
    'some value'
    >>> cache['item4'] = 'This will push item 2 out of the cache'
    >>> cache['item2']
    Traceback (most recent call last):
      File "<stdin>", line 1, in <module>
      File "lrucache.py", line 48, in __getitem__
        return super().__getitem__(key)
    KeyError: 'item2'
    >>> cache['item5'] = 'This will push item3 out of the cache'
    >>> cache['item3']
    Traceback (most recent call last):
      File "<stdin>", line 1, in <module>
      File "lrucache.py", line 48, in __getitem__
        return super().__getitem__(key)
    KeyError: 'item3'
    >>> cache['item1']
    'some value'
    """

    def __init__(self, max_size, *args, **kwargs):
        self._ordered_keys = []
        self._max_size = max_size
        self._dict = dict()
        for k, v in dict(*args, **kwargs).items():
            self[k] = v

    def __getitem__(self, key):
        value = self._dict[key]
        self._ordered_keys.remove(key)
        self._ordered_keys.insert(0, key)
        return value

    def __setitem__(self, key, value):
        if key in self:
            self._ordered_keys.remove(key)
        self._ordered_keys.insert(0, key)

        self._dict[key] = value

        if len(self) > self._max_size:
            item = self._ordered_keys.pop()
            del self._dict[item]

    def __delitem__(self, key):
        del self._dict[key]
        self._ordered_keys.remove(key)

    def __len__(self):
        return len(self._dict)

    def __iter__(self):
        return iter(self._dict)

    def __repr__(self):
        return repr(self._dict)



# -*- coding: utf-8 -*-

# Automatically generated - don't edit.
# Use `python setup.py build_ui` to update it.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_AboutDialog(object):
    def setupUi(self, AboutDialog):
        AboutDialog.setObjectName("AboutDialog")
        AboutDialog.resize(680, 480)
        AboutDialog.setMinimumSize(QtCore.QSize(400, 300))
        self.vboxlayout = QtWidgets.QVBoxLayout(AboutDialog)
        self.vboxlayout.setContentsMargins(0, 0, 0, 0)
        self.vboxlayout.setSpacing(6)
        self.vboxlayout.setObjectName("vboxlayout")
        self.header = QtWidgets.QHBoxLayout()
        self.header.setContentsMargins(9, 9, 9, 0)
        self.header.setObjectName("header")
        spacerItem = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.header.addItem(spacerItem)
        self.logo = QtWidgets.QLabel(AboutDialog)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.logo.sizePolicy().hasHeightForWidth())
        self.logo.setSizePolicy(sizePolicy)
        self.logo.setMinimumSize(QtCore.QSize(48, 48))
        self.logo.setMaximumSize(QtCore.QSize(48, 48))
        self.logo.setPixmap(QtGui.QPixmap(":/images/128x128/org.musicbrainz.Picard.png"))
        self.logo.setScaledContents(True)
        self.logo.setObjectName("logo")
        self.header.addWidget(self.logo)
        self.app_name = QtWidgets.QLabel(AboutDialog)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.app_name.sizePolicy().hasHeightForWidth())
        self.app_name.setSizePolicy(sizePolicy)
        font = QtGui.QFont()
        font.setPointSize(20)
        font.setBold(True)
        font.setWeight(75)
        self.app_name.setFont(font)
        self.app_name.setTextFormat(QtCore.Qt.PlainText)
        self.app_name.setObjectName("app_name")
        self.header.addWidget(self.app_name)
        spacerItem1 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.header.addItem(spacerItem1)
        self.vboxlayout.addLayout(self.header)
        self.scrollArea = QtWidgets.QScrollArea(AboutDialog)
        self.scrollArea.setFrameShape(QtWidgets.QFrame.NoFrame)
        self.scrollArea.setFrameShadow(QtWidgets.QFrame.Plain)
        self.scrollArea.setLineWidth(0)
        self.scrollArea.setHorizontalScrollBarPolicy(QtCore.Qt.ScrollBarAlwaysOff)
        self.scrollArea.setWidgetResizable(True)
        self.scrollArea.setAlignment(QtCore.Qt.AlignLeading|QtCore.Qt.AlignLeft|QtCore.Qt.AlignTop)
        self.scrollArea.setObjectName("scrollArea")
        self.scrollAreaWidgetContents = QtWidgets.QWidget()
        self.scrollAreaWidgetContents.setGeometry(QtCore.QRect(0, 0, 680, 416))
        self.scrollAreaWidgetContents.setObjectName("scrollAreaWidgetContents")
        self.verticalLayout = QtWidgets.QVBoxLayout(self.scrollAreaWidgetContents)
        self.verticalLayout.setContentsMargins(9, 0, 9, 9)
        self.verticalLayout.setSpacing(6)
        self.verticalLayout.setObjectName("verticalLayout")
        self.label = QtWidgets.QLabel(self.scrollAreaWidgetContents)
        self.label.setText("")
        self.label.setAlignment(QtCore.Qt.AlignHCenter|QtCore.Qt.AlignTop)
        self.label.setWordWrap(True)
        self.label.setTextInteractionFlags(QtCore.Qt.LinksAccessibleByMouse|QtCore.Qt.TextSelectableByKeyboard|QtCore.Qt.TextSelectableByMouse)
        self.label.setObjectName("label")
        self.verticalLayout.addWidget(self.label)
        self.scrollArea.setWidget(self.scrollAreaWidgetContents)
        self.vboxlayout.addWidget(self.scrollArea)

        self.retranslateUi(AboutDialog)
        QtCore.QMetaObject.connectSlotsByName(AboutDialog)

    def retranslateUi(self, AboutDialog):
        _translate = QtCore.QCoreApplication.translate
        AboutDialog.setWindowTitle(_("About Picard"))
        self.app_name.setText(_("MusicBrainz Picard"))



# -*- coding: utf-8 -*-
#
# Picard, the next-generation MusicBrainz tagger
#
# Copyright (C) 2017 David Mandelberg
# Copyright (C) 2017-2018 Sambhav Kothari
# Copyright (C) 2017-2023 Laurent Monin
# Copyright (C) 2018-2023 Philipp Wolfer
# Copyright (C) 2019 Michael Wiencek
# Copyright (C) 2020 dukeyin
# Copyright (C) 2020, 2023 David Kellner
# Copyright (C) 2021 Bob Swift
# Copyright (C) 2021 Vladislav Karbovskii
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.


from types import SimpleNamespace

from picard import log
from picard.config import get_config
from picard.const import (
    ALIAS_TYPE_ARTIST_NAME_ID,
    ALIAS_TYPE_LEGAL_NAME_ID,
    RELEASE_FORMATS,
)
from picard.util import (
    format_time,
    linear_combination_of_weights,
    parse_amazon_url,
    translate_from_sortname,
)
from picard.util.script_detector_weighted import detect_script_weighted


_ARTIST_REL_TYPES = {
    'arranger': 'arranger',
    'audio': 'engineer',
    'chorus master': 'performer:chorus master',
    'composer': 'composer',
    'concertmaster': 'performer:concertmaster',
    'conductor': 'conductor',
    'engineer': 'engineer',
    'instrument arranger': 'arranger',
    'librettist': 'lyricist',
    'live sound': 'engineer',
    'lyricist': 'lyricist',
    # 'mastering': 'engineer',
    'mix-DJ': 'djmixer',
    'mix': 'mixer',
    'orchestrator': 'arranger',
    'performing orchestra': 'performer:orchestra',
    'producer': 'producer',
    # 'recording': 'engineer',
    'remixer': 'remixer',
    'sound': 'engineer',
    'audio director': 'director',
    'video director': 'director',
    'vocal arranger': 'arranger',
    'writer': 'writer',
}

_TRACK_TO_METADATA = {
    'number': '~musicbrainz_tracknumber',
    'position': 'tracknumber',
    'title': 'title',
}

_MEDIUM_TO_METADATA = {
    'format': 'media',
    'position': 'discnumber',
    'title': 'discsubtitle',
}

_RECORDING_TO_METADATA = {
    'disambiguation': '~recordingcomment',
    'first-release-date': '~recording_firstreleasedate',
    'title': 'title',
}

_RELEASE_TO_METADATA = {
    'annotation': '~releaseannotation',
    'asin': 'asin',
    'barcode': 'barcode',
    'country': 'releasecountry',
    'date': 'date',
    'disambiguation': '~releasecomment',
    'title': 'album',
}

_ARTIST_TO_METADATA = {
    'gender': 'gender',
    'name': 'name',
    'type': 'type',
}

_RELEASE_GROUP_TO_METADATA = {
    'disambiguation': '~releasegroupcomment',
    'first-release-date': '~releasegroup_firstreleasedate',
    'title': '~releasegroup',
}

_PREFIX_ATTRS = {'guest', 'additional', 'minor', 'solo'}
_BLANK_SPECIAL_RELTYPES = {'vocal': 'vocals'}


def _parse_attributes(attrs, reltype, attr_credits):
    prefixes = []
    nouns = []
    for attr in attrs:
        if attr in attr_credits:
            attr = attr_credits[attr]
        if attr in _PREFIX_ATTRS:
            prefixes.append(attr)
        else:
            nouns.append(attr)
    if len(nouns) > 1:
        result = "%s and %s" % (", ".join(nouns[:-1]), nouns[-1:][0])
    elif len(nouns) == 1:
        result = nouns[0]
    else:
        result = _BLANK_SPECIAL_RELTYPES.get(reltype, "")
    prefix = " ".join(prefixes)
    return " ".join([prefix, result]).strip()


def _relation_attributes(relation):
    try:
        return tuple(relation['attributes'])
    except KeyError:
        return tuple()


def _relations_to_metadata_target_type_artist(relation, m, context):
    artist = relation['artist']
    translated_name, sort_name = _translate_artist_node(artist, config=context.config)
    has_translation = (translated_name != artist['name'])
    if not has_translation and context.use_credited_as and 'target-credit' in relation:
        credited_as = relation['target-credit']
        if credited_as:
            translated_name = credited_as
    reltype = relation['type']
    attribs = _relation_attributes(relation)
    if reltype in {'vocal', 'instrument', 'performer'}:
        if context.use_instrument_credits:
            attr_credits = relation.get('attribute-credits', {})
        else:
            attr_credits = {}
        name = 'performer:' + _parse_attributes(attribs, reltype, attr_credits)
    elif reltype == 'mix-DJ' and attribs:
        if not hasattr(m, '_djmix_ars'):
            m._djmix_ars = {}
        for attr in attribs:
            m._djmix_ars.setdefault(attr.split()[1], []).append(translated_name)
        return
    else:
        try:
            name = _ARTIST_REL_TYPES[reltype]
        except KeyError:
            return
    if context.instrumental and name == 'lyricist':
        return
    m.add_unique(name, translated_name)
    if name == 'composer':
        m.add_unique('composersort', sort_name)
    elif name == 'lyricist':
        m.add_unique('~lyricistsort', sort_name)
    elif name == 'writer':
        m.add_unique('~writersort', sort_name)


def _relations_to_metadata_target_type_work(relation, m, context):
    if relation['type'] == 'performance':
        performance_attributes = _relation_attributes(relation)
        for attribute in performance_attributes:
            m.add_unique('~performance_attributes', attribute)
        instrumental = 'instrumental' in performance_attributes
        work_to_metadata(relation['work'], m, instrumental)


def _relations_to_metadata_target_type_url(relation, m, context):
    if relation['type'] == 'amazon asin' and 'asin' not in m:
        amz = parse_amazon_url(relation['url']['resource'])
        if amz is not None:
            m['asin'] = amz['asin']
    elif relation['type'] == 'license':
        url = relation['url']['resource']
        m.add('license', url)


def _relations_to_metadata_target_type_series(relation, m, context):
    if relation['type'] == 'part of':
        entity = context.entity
        series = relation['series']
        var_prefix = f'~{entity}_' if entity else '~'
        name = f'{var_prefix}series'
        mbid = f'{var_prefix}seriesid'
        comment = f'{var_prefix}seriescomment'
        number = f'{var_prefix}seriesnumber'
        if not context.metadata_was_cleared['series']:
            # Clear related metadata first to prevent accumulation
            # of identical value, see PICARD-2700 issue
            m.unset(name)
            m.unset(mbid)
            m.unset(comment)
            m.unset(number)
            # That's to ensure it is done only once
            context.metadata_was_cleared['series'] = True
        m.add(name, series['name'])
        m.add(mbid, series['id'])
        m.add(comment, series['disambiguation'])
        m.add(number, relation['attribute-values'].get('number', ''))


class RelFunc(SimpleNamespace):
    clear_metadata_first = False
    func = None


_RELATIONS_TO_METADATA_TARGET_TYPE_FUNC = {
    'artist': RelFunc(func=_relations_to_metadata_target_type_artist),
    'series': RelFunc(
        func=_relations_to_metadata_target_type_series,
        clear_metadata_first=True
    ),
    'url': RelFunc(func=_relations_to_metadata_target_type_url),
    'work': RelFunc(func=_relations_to_metadata_target_type_work),
}


def _relations_to_metadata(relations, m, instrumental=False, config=None, entity=None):
    config = config or get_config()
    context = SimpleNamespace(
        config=config,
        entity=entity,
        instrumental=instrumental,
        use_credited_as=not config.setting['standardize_artists'],
        use_instrument_credits=not config.setting['standardize_instruments'],
        metadata_was_cleared=dict(),
    )
    for relation in relations:
        target = relation['target-type']
        if target in _RELATIONS_TO_METADATA_TARGET_TYPE_FUNC:
            relfunc = _RELATIONS_TO_METADATA_TARGET_TYPE_FUNC[target]
            if target not in context.metadata_was_cleared:
                context.metadata_was_cleared[target] = not relfunc.clear_metadata_first
            relfunc.func(relation, m, context)


def _locales_from_aliases(aliases):
    def check_higher_score(locale_dict, locale, score):
        return locale not in locale_dict or score > locale_dict[locale][0]

    full_locales = {}
    root_locales = {}
    for alias in aliases:
        if not alias.get('primary'):
            continue
        full_locale = alias.get('locale')
        if not full_locale:
            continue
        root_locale = full_locale.split('_')[0]
        full_parts = []
        root_parts = []
        score = 0.8
        full_parts.append((score, 5))
        if '_' in full_locale:
            score = 0.4
        root_parts.append((score, 5))
        type_id = alias.get('type-id')
        if type_id == ALIAS_TYPE_ARTIST_NAME_ID:
            score = 0.8
        elif type_id == ALIAS_TYPE_LEGAL_NAME_ID:
            score = 0.5
        else:
            # as 2014/09/19, only Artist or Legal names should have the
            # Primary flag
            score = 0.0
        full_parts.append((score, 5))
        root_parts.append((score, 5))
        comb = linear_combination_of_weights(full_parts)
        if check_higher_score(full_locales, full_locale, comb):
            full_locales[full_locale] = (comb, (alias['name'], alias['sort-name']))
        comb = linear_combination_of_weights(root_parts)
        if check_higher_score(root_locales, root_locale, comb):
            root_locales[root_locale] = (comb, (alias['name'], alias['sort-name']))

    return full_locales, root_locales


def _translate_artist_node(node, config=None):
    config = config or get_config()
    translated_name, sort_name = None, None
    if config.setting['translate_artist_names']:
        if config.setting['translate_artist_names_script_exception']:
            log_text = 'Script alpha characters found in "{0}": '.format(node['name'],)
            detected_scripts = detect_script_weighted(node['name'])
            if detected_scripts:
                log_text += "; ".join(
                    "{0} ({1:.1f}%)".format(scr_id, detected_scripts[scr_id] * 100)
                    for scr_id in detected_scripts
                )
            else:
                log_text += "None"
            log.debug(log_text)
            if detected_scripts:
                script_exceptions = config.setting['script_exceptions']
                if script_exceptions:
                    log_text = " found in selected scripts: " + "; ".join(
                        "{0} ({1}%)".format(scr[0], scr[1])
                        for scr in script_exceptions
                    )
                    for script_id, script_weighting in script_exceptions:
                        if script_id not in detected_scripts:
                            continue
                        if detected_scripts[script_id] >= script_weighting / 100:
                            log.debug("Match" + log_text)
                            return node['name'], node['sort-name']
                    log.debug("No match" + log_text)
                else:
                    log.warning("No scripts selected for translation exception match check.")

        # Prepare dictionaries of available locale aliases
        if 'aliases' in node:
            full_locales, root_locales = _locales_from_aliases(node['aliases'])

            # First pass to match full locale if available
            for locale in config.setting['artist_locales']:
                if locale in full_locales:
                    return full_locales[locale][1]

            # Second pass to match root locale if available
            for locale in config.setting['artist_locales']:
                lang = locale.split('_')[0]
                if lang in root_locales:
                    return root_locales[lang][1]

        # No matches found in available alias locales
        sort_name = node['sort-name']
        translated_name = translate_from_sortname(node['name'] or '', sort_name)
    else:
        translated_name, sort_name = node['name'], node['sort-name']
    return (translated_name, sort_name)


def artist_credit_from_node(node):
    artist_name = ''
    artist_sort_name = ''
    artist_names = []
    artist_sort_names = []
    config = get_config()
    use_credited_as = not config.setting['standardize_artists']
    for artist_info in node:
        artist = artist_info['artist']
        translated_name, sort_name = _translate_artist_node(artist, config=config)
        has_translation = (translated_name != artist['name'])
        if has_translation:
            name = translated_name
        elif use_credited_as and 'name' in artist_info:
            name = artist_info['name']
        else:
            name = artist['name']
        artist_name += name
        artist_sort_name += sort_name or ''
        artist_names.append(name)
        artist_sort_names.append(sort_name or '')
        if 'joinphrase' in artist_info:
            artist_name += artist_info['joinphrase'] or ''
            artist_sort_name += artist_info['joinphrase'] or ''
    return (artist_name, artist_sort_name, artist_names, artist_sort_names)


def artist_credit_to_metadata(node, m, release=False):
    ids = [n['artist']['id'] for n in node]
    artist_name, artist_sort_name, artist_names, artist_sort_names = artist_credit_from_node(node)
    if release:
        m['musicbrainz_albumartistid'] = ids
        m['albumartist'] = artist_name
        m['albumartistsort'] = artist_sort_name
        m['~albumartists'] = artist_names
        m['~albumartists_sort'] = artist_sort_names
    else:
        m['musicbrainz_artistid'] = ids
        m['artist'] = artist_name
        m['artistsort'] = artist_sort_name
        m['artists'] = artist_names
        m['~artists_sort'] = artist_sort_names


def _release_event_iter(node):
    if 'release-events' in node:
        yield from node['release-events']


def _country_from_release_event(release_event):
    try:
        return release_event['area']['iso-3166-1-codes'][0]
    # TypeError in case object is None
    except (KeyError, IndexError, TypeError):
        pass
    return None


def countries_from_node(node):
    countries = []
    for release_event in _release_event_iter(node):
        country_code = _country_from_release_event(release_event)
        if country_code:
            countries.append(country_code)
    return sorted(countries)


def release_dates_and_countries_from_node(node):
    dates = []
    countries = []
    for release_event in _release_event_iter(node):
        dates.append(release_event['date'] or '')
        country_code = _country_from_release_event(release_event)
        if country_code:
            countries.append(country_code)
    return dates, countries


def label_info_from_node(node):
    labels = []
    catalog_numbers = []
    for label_info in node:
        if 'label' in label_info and label_info['label'] and 'name' in label_info['label']:
            label = label_info['label']['name']
            if label and label not in labels:
                labels.append(label)
        if 'catalog-number' in label_info:
            cat_num = label_info['catalog-number']
            if cat_num and cat_num not in catalog_numbers:
                catalog_numbers.append(cat_num)
    return (labels, catalog_numbers)


def media_formats_from_node(node):
    formats_count = {}
    formats_order = []
    for medium in node:
        text = medium.get('format', "(unknown)") or "(unknown)"
        if text in formats_count:
            formats_count[text] += 1
        else:
            formats_count[text] = 1
            formats_order.append(text)
    formats = []
    for medium_format in formats_order:
        count = formats_count[medium_format]
        medium_format = RELEASE_FORMATS.get(medium_format, medium_format)
        if count > 1:
            medium_format = str(count) + "×" + medium_format
        formats.append(medium_format)
    return " + ".join(formats)


def _node_skip_empty_iter(node):
    for key, value in node.items():
        if value or value == 0:
            yield key, value


def track_to_metadata(node, track):
    m = track.metadata
    recording_to_metadata(node['recording'], m, track)
    m.add_unique('musicbrainz_trackid', node['id'])
    # overwrite with data we have on the track
    for key, value in _node_skip_empty_iter(node):
        if key in _TRACK_TO_METADATA:
            m[_TRACK_TO_METADATA[key]] = value
        elif key == 'length' and value:
            m.length = value
        elif key == 'artist-credit':
            artist_credit_to_metadata(value, m)
    if m.length:
        m['~length'] = format_time(m.length)


def recording_to_metadata(node, m, track=None):
    m.length = 0
    m.add_unique('musicbrainz_recordingid', node['id'])
    config = get_config()
    for key, value in _node_skip_empty_iter(node):
        if key in _RECORDING_TO_METADATA:
            m[_RECORDING_TO_METADATA[key]] = value
        elif key == 'user-rating':
            m['~rating'] = value['value']
        elif key == 'length':
            m.length = value
        elif key == 'artist-credit':
            artist_credit_to_metadata(value, m)
            # set tags from artists
            if track:
                for credit in value:
                    artist = credit['artist']
                    artist_obj = track.append_track_artist(artist['id'])
                    add_genres_from_node(artist, artist_obj)
        elif key == 'relations':
            _relations_to_metadata(value, m, config=config, entity='recording')
        elif key == 'isrcs':
            add_isrcs_to_metadata(value, m)
        elif key == 'video' and value:
            m['~video'] = '1'
    add_genres_from_node(node, track)
    if m['title']:
        m['~recordingtitle'] = m['title']
    if m.length:
        m['~length'] = format_time(m.length)


def work_to_metadata(work, m, instrumental=False):
    m.add_unique('musicbrainz_workid', work['id'])
    if instrumental:
        m.add_unique('language', 'zxx')  # no lyrics
    elif 'languages' in work:
        for language in work['languages']:
            m.add_unique('language', language)
    elif 'language' in work:
        m.add_unique('language', work['language'])
    if 'title' in work:
        m.add_unique('work', work['title'])
    if 'disambiguation' in work:
        m.add_unique('~workcomment', work['disambiguation'])
    if 'relations' in work:
        _relations_to_metadata(work['relations'], m, instrumental, entity='work')


def medium_to_metadata(node, m):
    for key, value in _node_skip_empty_iter(node):
        if key in _MEDIUM_TO_METADATA:
            m[_MEDIUM_TO_METADATA[key]] = value
    totaltracks = node.get('track-count', 0)
    if node.get('pregap'):
        totaltracks += 1
    if totaltracks:
        m['totaltracks'] = totaltracks


def artist_to_metadata(node, m):
    """Make meatadata dict from a JSON 'artist' node."""
    m.add_unique('musicbrainz_artistid', node['id'])
    for key, value in _node_skip_empty_iter(node):
        if key in _ARTIST_TO_METADATA:
            m[_ARTIST_TO_METADATA[key]] = value
        elif key == 'area':
            m['area'] = value['name']
        elif key == 'life-span':
            if 'begin' in value:
                m['begindate'] = value['begin']
            if 'ended' in value:
                ended = value['ended']
                if ended and 'end' in value:
                    m['enddate'] = value['end']
        elif key == 'begin-area':
            m['beginarea'] = value['name']
        elif key == 'end-area':
            m['endarea'] = value['name']


def release_to_metadata(node, m, album=None):
    """Make metadata dict from a JSON 'release' node."""
    config = get_config()
    m.add_unique('musicbrainz_albumid', node['id'])
    for key, value in _node_skip_empty_iter(node):
        if key in _RELEASE_TO_METADATA:
            m[_RELEASE_TO_METADATA[key]] = value
        elif key == 'status':
            m['releasestatus'] = value.lower()
        elif key == 'artist-credit':
            artist_credit_to_metadata(value, m, release=True)
            # set tags from artists
            if album is not None:
                for credit in value:
                    artist = credit['artist']
                    artist_obj = album.append_album_artist(artist['id'])
                    add_genres_from_node(artist, artist_obj)
        elif key == 'relations' and config.setting['release_ars']:
            _relations_to_metadata(value, m, config=config, entity='release')
        elif key == 'label-info':
            m['label'], m['catalognumber'] = label_info_from_node(value)
        elif key == 'text-representation':
            if 'language' in value:
                m['~releaselanguage'] = value['language']
            if 'script' in value:
                m['script'] = value['script']
    m['~releasecountries'] = release_countries = countries_from_node(node)
    # The MB web service returns the first release country in the country tag.
    # If the user has configured preferred release countries, use the first one
    # if it is one in the complete list of release countries.
    for country in config.setting['preferred_release_countries']:
        if country in release_countries:
            m['releasecountry'] = country
            break
    add_genres_from_node(node, album)


def release_group_to_metadata(node, m, release_group=None):
    """Make metadata dict from a JSON 'release-group' node taken from inside a 'release' node."""
    config = get_config()
    m.add_unique('musicbrainz_releasegroupid', node['id'])
    for key, value in _node_skip_empty_iter(node):
        if key in _RELEASE_GROUP_TO_METADATA:
            m[_RELEASE_GROUP_TO_METADATA[key]] = value
        elif key == 'primary-type':
            m['~primaryreleasetype'] = value.lower()
        elif key == 'secondary-types':
            add_secondary_release_types(value, m)
        elif key == 'relations' and config.setting['release_ars']:
            _relations_to_metadata(value, m, config=config, entity='releasegroup')
    add_genres_from_node(node, release_group)
    if m['~releasegroup_firstreleasedate']:
        m['originaldate'] = m['~releasegroup_firstreleasedate']
        m['originalyear'] = m['originaldate'][:4]
    m['releasetype'] = m.getall('~primaryreleasetype') + m.getall('~secondaryreleasetype')


def add_secondary_release_types(node, m):
    for secondary_type in node:
        m.add_unique('~secondaryreleasetype', secondary_type.lower())


def add_genres_from_node(node, obj):
    if obj is None:
        return
    if 'genres' in node:
        add_genres(node['genres'], obj)
    if 'tags' in node:
        add_genres(node['tags'], obj)
    if 'user-genres' in node:
        add_user_genres(node['user-genres'], obj)
    if 'user-tags' in node:
        add_user_genres(node['user-tags'], obj)


def add_genres(node, obj):
    for tag in node:
        obj.add_genre(tag['name'], tag['count'])


def add_user_genres(node, obj):
    for tag in node:
        obj.add_genre(tag['name'], 1)


def add_isrcs_to_metadata(node, metadata):
    for isrc in node:
        metadata.add('isrc', isrc)


def get_score(node):
    """Returns the score attribute for a node.
    The score is expected to be an integer between 0 and 100, it is returned as
    a value between 0.0 and 1.0. If there is no score attribute or it has an
    invalid value 1.0 will be returned.
    """
    try:
        return int(node.get('score', 100)) / 100
    except (TypeError, ValueError):
        return 1.0



# -*- coding: utf-8 -*-
#
# Picard, the next-generation MusicBrainz tagger
#
# Copyright (C) 2006-2009, 2011-2012 Lukáš Lalinský
# Copyright (C) 2008-2011, 2014, 2018-2021, 2023 Philipp Wolfer
# Copyright (C) 2009 Carlin Mangar
# Copyright (C) 2011-2012 Johannes Weißl
# Copyright (C) 2011-2014 Michael Wiencek
# Copyright (C) 2011-2014 Wieland Hoffmann
# Copyright (C) 2013 Calvin Walton
# Copyright (C) 2013-2014, 2017-2021 Laurent Monin
# Copyright (C) 2013-2015, 2017, 2021 Sophist-UK
# Copyright (C) 2015 Frederik “Freso” S. Olesen
# Copyright (C) 2016 Christoph Reiter
# Copyright (C) 2016-2018 Sambhav Kothari
# Copyright (C) 2017 tungol
# Copyright (C) 2019 Zenara Daley
# Copyright (C) 2023 certuna
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.


from collections import Counter
from enum import IntEnum
import re
from urllib.parse import urlparse

from mutagen import id3
import mutagen.aiff
import mutagen.apev2
import mutagen.dsf
import mutagen.mp3
import mutagen.trueaudio

from picard import log
from picard.config import get_config
from picard.coverart.image import (
    CoverArtImageError,
    TagCoverArtImage,
)
from picard.coverart.utils import types_from_id3
from picard.file import File
from picard.formats.mutagenext import (
    compatid3,
    delall_ci,
)
from picard.metadata import Metadata
from picard.util import (
    encode_filename,
    sanitize_date,
)
from picard.util.tags import parse_comment_tag


UNSUPPORTED_TAGS = {'r128_album_gain', 'r128_track_gain'}

id3.GRP1 = compatid3.GRP1


class Id3Encoding(IntEnum):
    LATIN1 = 0
    UTF16 = 1
    UTF16BE = 2
    UTF8 = 3

    def from_config(id3v2_encoding):
        return {
            'utf-8': Id3Encoding.UTF8,
            'utf-16': Id3Encoding.UTF16
        }.get(id3v2_encoding, Id3Encoding.LATIN1)


def id3text(text, encoding):
    """Returns a string which only contains code points which can
    be encododed with the given numeric id3 encoding.
    """

    if encoding == Id3Encoding.LATIN1:
        return text.encode('latin1', 'replace').decode('latin1')
    return text


def _remove_people_with_role(tags, frames, role):
    for frame in tags.values():
        if frame.FrameID in frames:
            for people in list(frame.people):
                if people[0] == role:
                    frame.people.remove(people)


class ID3File(File):

    """Generic ID3-based file."""
    _IsMP3 = False

    __upgrade = {
        'XSOP': 'TSOP',
        'TXXX:ALBUMARTISTSORT': 'TSO2',
        'TXXX:COMPOSERSORT': 'TSOC',
        'TXXX:mood': 'TMOO',
        'TXXX:RELEASEDATE': 'TDRL',
    }

    __translate = {
        # In same sequence as defined at http://id3.org/id3v2.4.0-frames
        # 'TIT1': 'grouping', # Depends on itunes_compatible_grouping
        'TIT2': 'title',
        'TIT3': 'subtitle',
        'TALB': 'album',
        'TSST': 'discsubtitle',
        'TSRC': 'isrc',
        'TPE1': 'artist',
        'TPE2': 'albumartist',
        'TPE3': 'conductor',
        'TPE4': 'remixer',
        'TEXT': 'lyricist',
        'TCOM': 'composer',
        'TENC': 'encodedby',
        'TBPM': 'bpm',
        'TKEY': 'key',
        'TLAN': 'language',
        'TCON': 'genre',
        'TMED': 'media',
        'TMOO': 'mood',
        'TCOP': 'copyright',
        'TPUB': 'label',
        'TDOR': 'originaldate',
        'TDRC': 'date',
        'TDRL': 'releasedate',
        'TSSE': 'encodersettings',
        'TSOA': 'albumsort',
        'TSOP': 'artistsort',
        'TSOT': 'titlesort',
        'WCOP': 'license',
        'WOAR': 'website',
        'COMM': 'comment',
        'TOAL': 'originalalbum',
        'TOPE': 'originalartist',
        'TOFN': 'originalfilename',

        # The following are informal iTunes extensions to id3v2:
        'TCMP': 'compilation',
        'TSOC': 'composersort',
        'TSO2': 'albumartistsort',
        'MVNM': 'movement'
    }
    __rtranslate = {v: k for k, v in __translate.items()}
    __translate['GRP1'] = 'grouping'  # Always read, but writing depends on itunes_compatible_grouping

    __translate_freetext = {
        'MusicBrainz Artist Id': 'musicbrainz_artistid',
        'MusicBrainz Album Id': 'musicbrainz_albumid',
        'MusicBrainz Album Artist Id': 'musicbrainz_albumartistid',
        'MusicBrainz Album Type': 'releasetype',
        'MusicBrainz Album Status': 'releasestatus',
        'MusicBrainz TRM Id': 'musicbrainz_trmid',
        'MusicBrainz Release Track Id': 'musicbrainz_trackid',
        'MusicBrainz Disc Id': 'musicbrainz_discid',
        'MusicBrainz Work Id': 'musicbrainz_workid',
        'MusicBrainz Release Group Id': 'musicbrainz_releasegroupid',
        'MusicBrainz Original Album Id': 'musicbrainz_originalalbumid',
        'MusicBrainz Original Artist Id': 'musicbrainz_originalartistid',
        'MusicBrainz Album Release Country': 'releasecountry',
        'MusicIP PUID': 'musicip_puid',
        'Acoustid Fingerprint': 'acoustid_fingerprint',
        'Acoustid Id': 'acoustid_id',
        'SCRIPT': 'script',
        'LICENSE': 'license',
        'CATALOGNUMBER': 'catalognumber',
        'BARCODE': 'barcode',
        'ASIN': 'asin',
        'MusicMagic Fingerprint': 'musicip_fingerprint',
        'ARTISTS': 'artists',
        'DIRECTOR': 'director',
        'WORK': 'work',
        'Writer': 'writer',
        'SHOWMOVEMENT': 'showmovement',
    }
    __rtranslate_freetext = {v: k for k, v in __translate_freetext.items()}
    __translate_freetext['writer'] = 'writer'  # For backward compatibility of case

    # Freetext fields that are loaded case-insensitive
    __rtranslate_freetext_ci = {
        'replaygain_album_gain': 'REPLAYGAIN_ALBUM_GAIN',
        'replaygain_album_peak': 'REPLAYGAIN_ALBUM_PEAK',
        'replaygain_album_range': 'REPLAYGAIN_ALBUM_RANGE',
        'replaygain_track_gain': 'REPLAYGAIN_TRACK_GAIN',
        'replaygain_track_peak': 'REPLAYGAIN_TRACK_PEAK',
        'replaygain_track_range': 'REPLAYGAIN_TRACK_RANGE',
        'replaygain_reference_loudness': 'REPLAYGAIN_REFERENCE_LOUDNESS',
    }
    __translate_freetext_ci = {b.lower(): a for a, b in __rtranslate_freetext_ci.items()}

    # Obsolete tag names which will still be loaded, but will get renamed on saving
    __rename_freetext = {
        'Artists': 'ARTISTS',
        'Work': 'WORK',
    }
    __rrename_freetext = {v: k for k, v in __rename_freetext.items()}

    _tipl_roles = {
        'engineer': 'engineer',
        'arranger': 'arranger',
        'producer': 'producer',
        'DJ-mix': 'djmixer',
        'mix': 'mixer',
    }
    _rtipl_roles = {v: k for k, v in _tipl_roles.items()}

    __other_supported_tags = ('discnumber', 'tracknumber',
                              'totaldiscs', 'totaltracks',
                              'movementnumber', 'movementtotal')
    __tag_re_parse = {
        'TRCK': re.compile(r'^(?P<tracknumber>\d+)(?:/(?P<totaltracks>\d+))?$'),
        'TPOS': re.compile(r'^(?P<discnumber>\d+)(?:/(?P<totaldiscs>\d+))?$'),
        'MVIN': re.compile(r'^(?P<movementnumber>\d+)(?:/(?P<movementtotal>\d+))?$')
    }

    def __init__(self, filename):
        super().__init__(filename)
        self.__casemap = {}

    def build_TXXX(self, encoding, desc, values):
        """Construct and return a TXXX frame."""
        # This is here so that plugins can customize the behavior of TXXX
        # frames in particular via subclassing.
        # discussion: https://github.com/metabrainz/picard/pull/634
        # discussion: https://github.com/metabrainz/picard/pull/635
        # Used in the plugin "Compatible TXXX frames"
        # PR: https://github.com/metabrainz/picard-plugins/pull/83
        return id3.TXXX(encoding=encoding, desc=desc, text=values)

    def _load(self, filename):
        log.debug("Loading file %r", filename)
        self.__casemap = {}
        file = self._get_file(encode_filename(filename))
        tags = file.tags or {}
        config = get_config()
        itunes_compatible = config.setting['itunes_compatible_grouping']
        rating_user_email = id3text(config.setting['rating_user_email'], Id3Encoding.LATIN1)
        rating_steps = config.setting['rating_steps']
        # upgrade custom 2.3 frames to 2.4
        for old, new in self.__upgrade.items():
            if old in tags and new not in tags:
                f = tags.pop(old)
                tags.add(getattr(id3, new)(encoding=f.encoding, text=f.text))
        metadata = Metadata()
        for frame in tags.values():
            frameid = frame.FrameID
            if frameid in self.__translate:
                name = self.__translate[frameid]
                if frameid.startswith('T') or frameid in {'GRP1', 'MVNM'}:
                    for text in frame.text:
                        if text:
                            metadata.add(name, text)
                elif frameid == 'COMM':
                    for text in frame.text:
                        if text:
                            if frame.lang == 'eng':
                                name = '%s:%s' % (name, frame.desc)
                            else:
                                name = '%s:%s:%s' % (name, frame.lang, frame.desc)
                            metadata.add(name, text)
                else:
                    metadata.add(name, frame)
            elif frameid == 'TIT1':
                name = 'work' if itunes_compatible else 'grouping'
                for text in frame.text:
                    if text:
                        metadata.add(name, text)
            elif frameid == 'TMCL':
                for role, name in frame.people:
                    if role == 'performer':
                        role = ''
                    if role:
                        metadata.add('performer:%s' % role, name)
                    else:
                        metadata.add('performer', name)
            elif frameid == 'TIPL':
                # If file is ID3v2.3, TIPL tag could contain TMCL
                # so we will test for TMCL values and add to TIPL if not TMCL
                for role, name in frame.people:
                    if role in self._tipl_roles and name:
                        metadata.add(self._tipl_roles[role], name)
                    else:
                        if role == 'performer':
                            role = ''
                        if role:
                            metadata.add('performer:%s' % role, name)
                        else:
                            metadata.add('performer', name)
            elif frameid == 'TXXX':
                name = frame.desc
                name_lower = name.lower()
                if name in self.__rename_freetext:
                    name = self.__rename_freetext[name]
                if name_lower in self.__translate_freetext_ci:
                    orig_name = name
                    name = self.__translate_freetext_ci[name_lower]
                    self.__casemap[name] = orig_name
                elif name in self.__translate_freetext:
                    name = self.__translate_freetext[name]
                elif ((name in self.__rtranslate)
                      != (name in self.__rtranslate_freetext)):
                    # If the desc of a TXXX frame conflicts with the name of a
                    # Picard tag, load it into ~id3:TXXX:desc rather than desc.
                    #
                    # This basically performs an XOR, making sure that 'name'
                    # is in __rtranslate or __rtranslate_freetext, but not
                    # both. (Being in both implies we support reading it both
                    # ways.) Currently, the only tag in both is license.
                    name = '~id3:TXXX:' + name
                for text in frame.text:
                    metadata.add(name, text)
            elif frameid == 'USLT':
                name = 'lyrics'
                if frame.desc:
                    name += ':%s' % frame.desc
                metadata.add(name, frame.text)
            elif frameid == 'UFID' and frame.owner == "http://musicbrainz.org":
                metadata['musicbrainz_recordingid'] = frame.data.decode('ascii', 'ignore')
            elif frameid in self.__tag_re_parse.keys():
                m = self.__tag_re_parse[frameid].search(frame.text[0])
                if m:
                    for name, value in m.groupdict().items():
                        if value is not None:
                            metadata[name] = value
                else:
                    log.error("Invalid %s value '%s' dropped in %r", frameid, frame.text[0], filename)
            elif frameid == 'APIC':
                try:
                    coverartimage = TagCoverArtImage(
                        file=filename,
                        tag=frameid,
                        types=types_from_id3(frame.type),
                        comment=frame.desc,
                        support_types=True,
                        data=frame.data,
                        id3_type=frame.type,
                    )
                except CoverArtImageError as e:
                    log.error("Cannot load image from %r: %s", filename, e)
                else:
                    metadata.images.append(coverartimage)
            elif frameid == 'POPM':
                # Rating in ID3 ranges from 0 to 255, normalize this to the range 0 to 5
                if frame.email == rating_user_email:
                    rating = int(round(frame.rating / 255.0 * (rating_steps - 1)))
                    metadata.add('~rating', rating)

        if 'date' in metadata:
            sanitized = sanitize_date(metadata.getall('date')[0])
            if sanitized:
                metadata['date'] = sanitized

        self._info(metadata, file)
        return metadata

    def _save(self, filename, metadata):
        """Save metadata to the file."""
        log.debug("Saving file %r", filename)
        tags = self._get_tags(filename)
        config = get_config()
        if config.setting['clear_existing_tags']:
            cover = tags.getall('APIC') if config.setting['preserve_images'] else None
            tags.clear()
            if cover:
                tags.setall('APIC', cover)
        images_to_save = list(metadata.images.to_be_saved_to_tags())
        if images_to_save:
            tags.delall('APIC')

        encoding = Id3Encoding.from_config(config.setting['id3v2_encoding'])

        if 'tracknumber' in metadata:
            if 'totaltracks' in metadata:
                text = '%s/%s' % (metadata['tracknumber'], metadata['totaltracks'])
            else:
                text = metadata['tracknumber']
            tags.add(id3.TRCK(encoding=Id3Encoding.LATIN1, text=id3text(text, Id3Encoding.LATIN1)))

        if 'discnumber' in metadata:
            if 'totaldiscs' in metadata:
                text = '%s/%s' % (metadata['discnumber'], metadata['totaldiscs'])
            else:
                text = metadata['discnumber']
            tags.add(id3.TPOS(encoding=Id3Encoding.LATIN1, text=id3text(text, Id3Encoding.LATIN1)))

        if 'movementnumber' in metadata:
            if 'movementtotal' in metadata:
                text = '%s/%s' % (metadata['movementnumber'], metadata['movementtotal'])
            else:
                text = metadata['movementnumber']
            tags.add(id3.MVIN(encoding=Id3Encoding.LATIN1, text=id3text(text, Id3Encoding.LATIN1)))

        # This is necessary because mutagens HashKey for APIC frames only
        # includes the FrameID (APIC) and description - it's basically
        # impossible to save two images, even of different types, without
        # any description.
        counters = Counter()
        for image in images_to_save:
            desc = desctag = image.comment
            if counters[desc] > 0:
                if desc:
                    desctag = "%s (%i)" % (desc, counters[desc])
                else:
                    desctag = "(%i)" % counters[desc]
            counters[desc] += 1
            tags.add(id3.APIC(encoding=Id3Encoding.LATIN1,
                              mime=image.mimetype,
                              type=image.id3_type,
                              desc=id3text(desctag, Id3Encoding.LATIN1),
                              data=image.data))

        tmcl = mutagen.id3.TMCL(encoding=encoding, people=[])
        tipl = mutagen.id3.TIPL(encoding=encoding, people=[])
        for name, values in metadata.rawitems():
            values = [id3text(v, encoding) for v in values]
            name = id3text(name, encoding)
            name_lower = name.lower()

            if not self.supports_tag(name):
                continue
            elif name == 'performer' or name.startswith('performer:'):
                if ':' in name:
                    role = name.split(':', 1)[1]
                else:
                    role = 'performer'
                for value in values:
                    if config.setting['write_id3v23']:
                        # TIPL will be upgraded to IPLS
                        tipl.people.append([role, value])
                    else:
                        tmcl.people.append([role, value])
            elif name == 'comment' or name.startswith('comment:'):
                (lang, desc) = parse_comment_tag(name)
                if desc.lower()[:4] == 'itun':
                    tags.delall('COMM:' + desc)
                    tags.add(id3.COMM(encoding=Id3Encoding.LATIN1, desc=desc, lang='eng', text=[v + '\x00' for v in values]))
                else:
                    tags.add(id3.COMM(encoding=encoding, desc=desc, lang=lang, text=values))
            elif name.startswith('lyrics:') or name == 'lyrics':
                if ':' in name:
                    desc = name.split(':', 1)[1]
                else:
                    desc = ''
                for value in values:
                    tags.add(id3.USLT(encoding=encoding, desc=desc, text=value))
            elif name in self._rtipl_roles:
                for value in values:
                    tipl.people.append([self._rtipl_roles[name], value])
            elif name == 'musicbrainz_recordingid':
                tags.add(id3.UFID(owner="http://musicbrainz.org", data=bytes(values[0], 'ascii')))
            elif name == '~rating':
                rating_user_email = id3text(config.setting['rating_user_email'], Id3Encoding.LATIN1)
                # Search for an existing POPM frame to get the current playcount
                for frame in tags.values():
                    if frame.FrameID == 'POPM' and frame.email == rating_user_email:
                        count = getattr(frame, 'count', 0)
                        break
                else:
                    count = 0

                # Convert rating to range between 0 and 255
                rating = int(round(float(values[0]) * 255 / (config.setting['rating_steps'] - 1)))
                tags.add(id3.POPM(email=rating_user_email, rating=rating, count=count))
            elif name == 'grouping':
                if config.setting['itunes_compatible_grouping']:
                    tags.add(id3.GRP1(encoding=encoding, text=values))
                else:
                    tags.add(id3.TIT1(encoding=encoding, text=values))
            elif name == 'work' and config.setting['itunes_compatible_grouping']:
                tags.add(id3.TIT1(encoding=encoding, text=values))
                tags.delall('TXXX:Work')
                tags.delall('TXXX:WORK')
            elif name in self.__rtranslate:
                frameid = self.__rtranslate[name]
                if frameid.startswith('W'):
                    valid_urls = all(all(urlparse(v)[:2]) for v in values)
                    if frameid == 'WCOP':
                        # Only add WCOP if there is only one license URL, otherwise use TXXX:LICENSE
                        if len(values) > 1 or not valid_urls:
                            tags.delall('WCOP')
                            tags.add(self.build_TXXX(encoding, self.__rtranslate_freetext[name], values))
                        else:
                            tags.delall('TXXX:' + self.__rtranslate_freetext[name])
                            tags.add(id3.WCOP(url=values[0]))
                    elif frameid == 'WOAR' and valid_urls:
                        tags.delall('WOAR')
                        for url in values:
                            tags.add(id3.WOAR(url=url))
                elif frameid.startswith('T') or frameid == 'MVNM':
                    if config.setting['write_id3v23']:
                        if frameid == 'TMOO':
                            tags.add(self.build_TXXX(encoding, 'mood', values))
                    # No need to care about the TMOO tag being added again as it is
                    # automatically deleted by Mutagen if id2v23 is selected
                        if frameid == 'TDRL':
                            tags.add(self.build_TXXX(encoding, 'RELEASEDATE', values))
                    tags.add(getattr(id3, frameid)(encoding=encoding, text=values))
                    if frameid == 'TSOA':
                        tags.delall('XSOA')
                    elif frameid == 'TSOP':
                        tags.delall('XSOP')
                    elif frameid == 'TSO2':
                        tags.delall('TXXX:ALBUMARTISTSORT')
            elif name_lower in self.__rtranslate_freetext_ci:
                if name_lower in self.__casemap:
                    description = self.__casemap[name_lower]
                else:
                    description = self.__rtranslate_freetext_ci[name_lower]
                delall_ci(tags, 'TXXX:' + description)
                tags.add(self.build_TXXX(encoding, description, values))
            elif name in self.__rtranslate_freetext:
                description = self.__rtranslate_freetext[name]
                if description in self.__rrename_freetext:
                    tags.delall('TXXX:' + self.__rrename_freetext[description])
                tags.add(self.build_TXXX(encoding, description, values))
            elif name.startswith('~id3:'):
                name = name[5:]
                if name.startswith('TXXX:'):
                    tags.add(self.build_TXXX(encoding, name[5:], values))
                else:
                    frameclass = getattr(id3, name[:4], None)
                    if frameclass:
                        tags.add(frameclass(encoding=encoding, text=values))
            # don't save private / already stored tags
            elif not name.startswith('~') and name not in self.__other_supported_tags:
                tags.add(self.build_TXXX(encoding, name, values))

        if tmcl.people:
            tags.add(tmcl)
        if tipl.people:
            tags.add(tipl)

        self._remove_deleted_tags(metadata, tags)

        self._save_tags(tags, encode_filename(filename))

        if self._IsMP3 and config.setting['remove_ape_from_mp3']:
            try:
                mutagen.apev2.delete(encode_filename(filename))
            except BaseException:
                pass

    def _remove_deleted_tags(self, metadata, tags):
        """Remove the tags from the file that were deleted in the UI"""
        config = get_config()

        for name in metadata.deleted_tags:
            real_name = self._get_tag_name(name)
            try:
                if name.startswith('performer:'):
                    role = name.split(':', 1)[1]
                    _remove_people_with_role(tags, ['TMCL', 'TIPL', 'IPLS'], role)
                elif name.startswith('comment:') or name == 'comment':
                    (lang, desc) = parse_comment_tag(name)
                    for key, frame in list(tags.items()):
                        if (frame.FrameID == 'COMM' and frame.desc == desc
                            and frame.lang == lang):
                            del tags[key]
                elif name.startswith('lyrics:') or name == 'lyrics':
                    if ':' in name:
                        desc = name.split(':', 1)[1]
                    else:
                        desc = ''
                    for key, frame in list(tags.items()):
                        if frame.FrameID == 'USLT' and frame.desc == desc:
                            del tags[key]
                elif name in self._rtipl_roles:
                    role = self._rtipl_roles[name]
                    _remove_people_with_role(tags, ['TIPL', 'IPLS'], role)
                elif name == 'musicbrainz_recordingid':
                    for key, frame in list(tags.items()):
                        if frame.FrameID == 'UFID' and frame.owner == "http://musicbrainz.org":
                            del tags[key]
                elif name == 'license':
                    tags.delall(real_name)
                    tags.delall('TXXX:' + self.__rtranslate_freetext[name])
                elif real_name == 'POPM':
                    rating_user_email = id3text(config.setting['rating_user_email'], Id3Encoding.LATIN1)
                    for key, frame in list(tags.items()):
                        if frame.FrameID == 'POPM' and frame.email == rating_user_email:
                            del tags[key]
                elif real_name in self.__translate:
                    tags.delall(real_name)
                elif name.lower() in self.__rtranslate_freetext_ci:
                    delall_ci(tags, 'TXXX:' + self.__rtranslate_freetext_ci[name.lower()])
                elif real_name in self.__translate_freetext:
                    tags.delall('TXXX:' + real_name)
                    if real_name in self.__rrename_freetext:
                        tags.delall('TXXX:' + self.__rrename_freetext[real_name])
                elif not name.startswith('~id3:') and name not in self.__other_supported_tags:
                    tags.delall('TXXX:' + name)
                elif name.startswith('~id3:'):
                    frameid = name[5:]
                    tags.delall(frameid)
                elif name in self.__other_supported_tags:
                    del tags[real_name]
            except KeyError:
                pass

    @classmethod
    def supports_tag(cls, name):
        return ((name and not name.startswith('~') and name not in UNSUPPORTED_TAGS)
                or name == '~rating'
                or name.startswith('~id3'))

    def _get_tag_name(self, name):
        if name in self.__rtranslate:
            return self.__rtranslate[name]
        elif name in self.__rtranslate_freetext:
            return self.__rtranslate_freetext[name]
        elif name == '~rating':
            return 'POPM'
        elif name == 'tracknumber':
            return 'TRCK'
        elif name == 'discnumber':
            return 'TPOS'
        elif name == 'movementnumber':
            return 'MVIN'
        else:
            return None

    def _get_file(self, filename):
        raise NotImplementedError()

    def _get_tags(self, filename):
        try:
            return compatid3.CompatID3(encode_filename(filename))
        except mutagen.id3.ID3NoHeaderError:
            return compatid3.CompatID3()

    def _save_tags(self, tags, filename):
        config = get_config()
        if config.setting['write_id3v1']:
            v1 = 2
        else:
            v1 = 0

        if config.setting['write_id3v23']:
            tags.update_to_v23()
            separator = config.setting['id3v23_join_with']
            tags.save(filename, v2_version=3, v1=v1, v23_sep=separator)
        else:
            tags.update_to_v24()
            tags.save(filename, v2_version=4, v1=v1)

    def format_specific_metadata(self, metadata, tag, settings=None):
        if not settings:
            settings = get_config().setting

        if not settings['write_id3v23']:
            return super().format_specific_metadata(metadata, tag, settings)

        values = metadata.getall(tag)
        if not values:
            return values

        if tag == 'originaldate':
            values = [v[:4] for v in values]
        elif tag == 'date':
            values = [(v[:4] if len(v) < 10 else v) for v in values]

        # If this is a multi-valued field, then it needs to be flattened,
        # unless it's TIPL or TMCL which can still be multi-valued.
        if (len(values) > 1 and tag not in ID3File._rtipl_roles
                and not tag.startswith('performer:')):
            join_with = settings['id3v23_join_with']
            values = [join_with.join(values)]

        return values


class MP3File(ID3File):

    """MP3 file."""
    EXTENSIONS = [".mp3", ".mp2", ".m2a"]
    NAME = "MPEG-1 Audio"
    _IsMP3 = True
    _File = mutagen.mp3.MP3

    def _get_file(self, filename):
        return self._File(filename, ID3=compatid3.CompatID3)

    def _info(self, metadata, file):
        super()._info(metadata, file)
        id3version = ''
        if file.tags is not None and file.info.layer == 3:
            id3version = ' - ID3v%d.%d' % (file.tags.version[0], file.tags.version[1])
        metadata['~format'] = 'MPEG-1 Layer %d%s' % (file.info.layer, id3version)


class TrueAudioFile(ID3File):

    """TTA file."""
    EXTENSIONS = [".tta"]
    NAME = "The True Audio"
    _File = mutagen.trueaudio.TrueAudio

    def _get_file(self, filename):
        return self._File(filename, ID3=compatid3.CompatID3)


class NonCompatID3File(ID3File):
    """Base class for ID3 files which do not support setting `compatid3.CompatID3`."""

    def _get_file(self, filename):
        return self._File(filename, known_frames=compatid3.known_frames)

    def _get_tags(self, filename):
        file = self._get_file(filename)
        if file.tags is None:
            file.add_tags()
        return file.tags

    def _save_tags(self, tags, filename):
        config = get_config()
        if config.setting['write_id3v23']:
            compatid3.update_to_v23(tags)
            separator = config.setting['id3v23_join_with']
            tags.save(filename, v2_version=3, v23_sep=separator)
        else:
            tags.update_to_v24()
            tags.save(filename, v2_version=4)


class DSFFile(NonCompatID3File):

    """DSF file."""
    EXTENSIONS = [".dsf"]
    NAME = "DSF"
    _File = mutagen.dsf.DSF


class AiffFile(NonCompatID3File):

    """AIFF file."""
    EXTENSIONS = [".aiff", ".aif", ".aifc"]
    NAME = "Audio Interchange File Format (AIFF)"
    _File = mutagen.aiff.AIFF


try:
    import mutagen.dsdiff

    class DSDIFFFile(NonCompatID3File):

        """DSF file."""
        EXTENSIONS = [".dff"]
        NAME = "DSDIFF"
        _File = mutagen.dsdiff.DSDIFF

except ImportError:
    DSDIFFFile = None



# -*- coding: utf-8 -*-
#
# Picard, the next-generation MusicBrainz tagger
#
# Copyright (C) 2008 Lukáš Lalinský
# Copyright (C) 2013, 2018-2021 Laurent Monin
# Copyright (C) 2017 Sambhav Kothari
# Copyright (C) 2018-2019, 2022 Philipp Wolfer
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.


"""Tom's lossless Audio Kompressor streams with APEv2 tags.

TAK is a lossless audio compressor developed by Thomas Becker.

For more information, see http://wiki.hydrogenaudio.org/index.php?title=TAK
and http://en.wikipedia.org/wiki/TAK_(audio_codec)
"""

__all__ = ["TAK", "Open", "delete"]

try:
    from mutagen.tak import (
        TAK,
        Open,
        TAKHeaderError,
        TAKInfo,
        delete,
    )

    native_tak = True

except ImportError:
    from mutagen import StreamInfo
    from mutagen.apev2 import (
        APEv2File,
        delete,
        error,
    )

    native_tak = False

    class TAKHeaderError(error):
        pass

    class TAKInfo(StreamInfo):

        """TAK stream information.

        Attributes:
          (none at the moment)
        """

        def __init__(self, fileobj):
            header = fileobj.read(4)
            if len(header) != 4 or not header.startswith(b"tBaK"):
                raise TAKHeaderError("not a TAK file")

        @staticmethod
        def pprint():
            return "Tom's lossless Audio Kompressor"

    class TAK(APEv2File):
        """TAK(filething)

        Arguments:
            filething (filething)

        Attributes:
            info (`TAKInfo`)
        """

        _Info = TAKInfo
        _mimes = ["audio/x-tak"]

        @staticmethod
        def score(filename, fileobj, header):
            return header.startswith(b"tBaK") + filename.lower().endswith(".tak")

    Open = TAK



# -*- coding: utf-8 -*-
#
# Picard, the next-generation MusicBrainz tagger
#
# Copyright (C) 2018 Yvan Rivière
# Copyright (C) 2018, 2020-2021 Laurent Monin
# Copyright (C) 2018, 2020-2021 Philipp Wolfer
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.


from functools import partial

from PyQt5 import QtWidgets

from picard import log
from picard.album import Album
from picard.cluster import (
    Cluster,
    ClusterList,
)
from picard.script import (
    ScriptError,
    ScriptParser,
)
from picard.track import Track
from picard.util import iter_unique


class ScriptsMenu(QtWidgets.QMenu):

    def __init__(self, scripts, *args):
        super().__init__(*args)

        for script in scripts:
            action = self.addAction(script[1])
            action.triggered.connect(partial(self._run_script, script))

    def _run_script(self, script):
        s_name = script[1]
        s_text = script[3]
        parser = ScriptParser()

        for obj in self._iter_unique_metadata_objects():
            try:
                parser.eval(s_text, obj.metadata)
                obj.update()
            except ScriptError as e:
                log.exception('Error running tagger script "%s" on object %r', s_name, obj)
                msg = N_('Script error in "%(script)s": %(message)s')
                mparms = {
                    'script': s_name,
                    'message': str(e),
                }
                self.tagger.window.set_statusbar_message(msg, mparms)

    def _iter_unique_metadata_objects(self):
        return iter_unique(self._iter_metadata_objects(self.tagger.window.selected_objects))

    def _iter_metadata_objects(self, objs):
        for obj in objs:
            if hasattr(obj, 'metadata') and not getattr(obj, 'special', False):
                yield obj
            if isinstance(obj, Cluster) or isinstance(obj, Track):
                yield from self._iter_metadata_objects(obj.iterfiles())
            elif isinstance(obj, ClusterList):
                yield from self._iter_metadata_objects(obj)
            elif isinstance(obj, Album):
                yield from self._iter_metadata_objects(obj.tracks)
                yield from self._iter_metadata_objects(obj.unmatched_files.iterfiles())



# -*- coding: utf-8 -*-
#
# Picard, the next-generation MusicBrainz tagger
#
# Copyright (C) 2006 Lukáš Lalinský
# Copyright (C) 2013, 2018, 2020-2021 Laurent Monin
# Copyright (C) 2013, 2020-2021 Philipp Wolfer
# Copyright (C) 2016-2017 Sambhav Kothari
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.


from picard.config import (
    BoolOption,
    IntOption,
    TextOption,
    get_config,
)

from picard.ui.options import (
    OptionsPage,
    register_options_page,
)
from picard.ui.ui_options_network import Ui_NetworkOptionsPage


class NetworkOptionsPage(OptionsPage):

    NAME = 'network'
    TITLE = N_("Network")
    PARENT = 'advanced'
    SORT_ORDER = 10
    ACTIVE = True
    HELP_URL = "/config/options_network.html"

    options = [
        BoolOption('setting', 'use_proxy', False),
        TextOption('setting', 'proxy_type', 'http'),
        TextOption('setting', 'proxy_server_host', ''),
        IntOption('setting', 'proxy_server_port', 80),
        TextOption('setting', 'proxy_username', ''),
        TextOption('setting', 'proxy_password', ''),
        BoolOption('setting', 'browser_integration', True),
        IntOption('setting', 'browser_integration_port', 8000),
        BoolOption('setting', 'browser_integration_localhost_only', True),
        IntOption('setting', 'network_transfer_timeout_seconds', 30),
    ]

    def __init__(self, parent=None):
        super().__init__(parent)
        self.ui = Ui_NetworkOptionsPage()
        self.ui.setupUi(self)

    def load(self):
        config = get_config()
        self.ui.web_proxy.setChecked(config.setting['use_proxy'])
        if config.setting['proxy_type'] == 'socks':
            self.ui.proxy_type_socks.setChecked(True)
        else:
            self.ui.proxy_type_http.setChecked(True)
        self.ui.server_host.setText(config.setting['proxy_server_host'])
        self.ui.server_port.setValue(config.setting['proxy_server_port'])
        self.ui.username.setText(config.setting['proxy_username'])
        self.ui.password.setText(config.setting['proxy_password'])
        self.ui.transfer_timeout.setValue(config.setting['network_transfer_timeout_seconds'])
        self.ui.browser_integration.setChecked(config.setting['browser_integration'])
        self.ui.browser_integration_port.setValue(config.setting['browser_integration_port'])
        self.ui.browser_integration_localhost_only.setChecked(
            config.setting['browser_integration_localhost_only'])

    def save(self):
        config = get_config()
        config.setting['use_proxy'] = self.ui.web_proxy.isChecked()
        if self.ui.proxy_type_socks.isChecked():
            config.setting['proxy_type'] = 'socks'
        else:
            config.setting['proxy_type'] = 'http'
        config.setting['proxy_server_host'] = self.ui.server_host.text()
        config.setting['proxy_server_port'] = self.ui.server_port.value()
        config.setting['proxy_username'] = self.ui.username.text()
        config.setting['proxy_password'] = self.ui.password.text()
        self.tagger.webservice.setup_proxy()
        transfer_timeout = self.ui.transfer_timeout.value()
        config.setting['network_transfer_timeout_seconds'] = transfer_timeout
        self.tagger.webservice.set_transfer_timeout(transfer_timeout)
        config.setting['browser_integration'] = self.ui.browser_integration.isChecked()
        config.setting['browser_integration_port'] = self.ui.browser_integration_port.value()
        config.setting['browser_integration_localhost_only'] = \
            self.ui.browser_integration_localhost_only.isChecked()
        self.tagger.update_browser_integration()


register_options_page(NetworkOptionsPage)



# -*- coding: utf-8 -*-
#
# Picard, the next-generation MusicBrainz tagger
#
# Copyright (C) 2007 Lukáš Lalinský
# Copyright (C) 2012-2013, 2017 Wieland Hoffmann
# Copyright (C) 2013 Michael Wiencek
# Copyright (C) 2016-2017 Sambhav Kothari
# Copyright (C) 2018, 2020-2021 Laurent Monin
# Copyright (C) 2018-2022 Philipp Wolfer
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.


from collections.abc import MutableMapping

import mutagen

from picard import log
from picard.config import get_config
from picard.file import File
from picard.formats.id3 import NonCompatID3File
from picard.metadata import Metadata


try:
    from mutagen._iff import assert_valid_chunk_id
    from mutagen._riff import RiffFile
    from mutagen._util import loadfile
    import mutagen.wave

    # See https://exiftool.org/TagNames/RIFF.html
    TRANSLATE_RIFF_INFO = {
        # Minimal, as e.g. supported by Windows Explorer,
        # Audacity and foobar2000
        'IART': 'artist',
        'ICMT': 'comment',
        'ICOP': 'copyright',
        'ICRD': 'date',
        'IGNR': 'genre',
        'INAM': 'title',
        'IPRD': 'album',
        'ITRK': 'tracknumber',

        # Extended, not well supported by other tools
        'ICNT': 'releasecountry',
        'IENC': 'encodedby',
        'IENG': 'engineer',
        'ILNG': 'language',
        'IMED': 'media',
        'IMUS': 'composer',
        'IPRO': 'producer',
        'IWRI': 'writer',
    }

    R_TRANSLATE_RIFF_INFO = {v: k for k, v in TRANSLATE_RIFF_INFO.items()}

    def translate_tag_to_riff_name(name):
        if name.startswith('comment:'):
            name = 'comment'
        return R_TRANSLATE_RIFF_INFO.get(name, None)

    class RiffListInfo(MutableMapping):
        """Allows loading / saving RIFF INFO tags from / to RIFF files.
        """

        def __init__(self, encoding='windows-1252'):
            self.encoding = encoding
            self.__tags = {}
            self.__deleted_tags = set()

        @loadfile()
        def load(self, filething):
            """Load the INFO tags from the file."""
            riff_file = RiffFile(filething.fileobj)
            info = self.__find_info_chunk(riff_file.root)
            if info:
                for tag in info.subchunks():
                    self.__tags[tag.id] = self.__decode_data(tag.read())

        @loadfile(writable=True)
        def save(self, filething):
            """Save the INFO tags to the file."""
            riff_file = RiffFile(filething.fileobj)
            info = self.__find_info_chunk(riff_file.root)
            if not info:
                info = riff_file.insert_chunk('LIST', b'INFO')
            for name, value in self.__tags.items():
                self.__save_tag_data(info, name, value)
            for name in self.__deleted_tags:
                self.__delete_tag(info, name)

        @loadfile(writable=True)
        def delete(self, filething):
            """Deletes the INFO chunk completely from the file."""
            riff_file = RiffFile(filething.fileobj)
            info = self.__find_info_chunk(riff_file.root)
            if info:
                info.delete()

        @staticmethod
        def __find_info_chunk(parent):
            for chunk in parent.subchunks():
                if chunk.id == 'LIST' and chunk.name == 'INFO':
                    return chunk
            return None

        @staticmethod
        def __find_subchunk(parent, name):
            for chunk in parent.subchunks():
                if chunk.id == name:
                    return chunk
            return None

        def __save_tag_data(self, info, name, value):
            data = self.__encode_data(value)
            chunk = self.__find_subchunk(info, name)
            if chunk:
                chunk.resize(len(data))
                chunk.write(data)
                return chunk
            else:
                return info.insert_chunk(name, data)

        def __delete_tag(self, info, name):
            chunk = self.__find_subchunk(info, name)
            if chunk:
                chunk.delete()

        @staticmethod
        def __decode_data(value):
            try:  # Always try first to decode as Unicode
                value = value.decode('utf-8')
            except UnicodeDecodeError:  # Fall back to Windows-1252 encoding
                value = value.decode('windows-1252', errors='replace')
            return value.rstrip('\0')

        def __encode_data(self, value):
            return value.encode(self.encoding, errors='replace') + b'\x00'

        def __contains__(self, name):
            return self.__tags.__contains__(name)

        def __getitem__(self, key):
            return self.__tags.get(key)

        def __setitem__(self, key, value):
            assert_valid_chunk_id(key)
            self.__tags[key] = value
            self.__deleted_tags.discard(key)

        def __delitem__(self, key):
            if key in self.__tags:
                del self.__tags[key]
            self.__deleted_tags.add(key)

        def __iter__(self):
            return iter(self.__tags)

        def __len__(self):
            return len(self.__tags)

        def __repr__(self):
            return repr(self.__tags)

        def __str__(self):
            return str(self.__tags)

    class WAVFile(NonCompatID3File):
        EXTENSIONS = [".wav"]
        NAME = "Microsoft WAVE"
        _File = mutagen.wave.WAVE

        def _info(self, metadata, file):
            super()._info(metadata, file)
            metadata['~format'] = self.NAME
            config = get_config()
            info = RiffListInfo(encoding=config.setting['wave_riff_info_encoding'])
            info.load(file.filename)
            for tag, value in info.items():
                if tag in TRANSLATE_RIFF_INFO:
                    name = TRANSLATE_RIFF_INFO[tag]
                    if name not in metadata:
                        metadata[name] = value

        def _save(self, filename, metadata):
            super()._save(filename, metadata)

            # Save RIFF LIST INFO
            config = get_config()
            if config.setting['write_wave_riff_info']:
                info = RiffListInfo(encoding=config.setting['wave_riff_info_encoding'])
                if config.setting['clear_existing_tags']:
                    info.delete(filename)
                for name, values in metadata.rawitems():
                    name = translate_tag_to_riff_name(name)
                    if name:
                        value = ", ".join(values)
                        info[name] = value
                for name in metadata.deleted_tags:
                    name = translate_tag_to_riff_name(name)
                    if name:
                        del info[name]
                info.save(filename)
            elif config.setting['remove_wave_riff_info']:
                info = RiffListInfo(encoding=config.setting['wave_riff_info_encoding'])
                info.delete(filename)

except ImportError:
    import wave

    class WAVFile(File):
        EXTENSIONS = [".wav"]
        NAME = "Microsoft WAVE"
        _File = None

        def _load(self, filename):
            log.debug("Loading file %r", filename)
            f = wave.open(filename, "rb")
            metadata = Metadata()
            metadata['~channels'] = f.getnchannels()
            metadata['~bits_per_sample'] = f.getsampwidth() * 8
            metadata['~sample_rate'] = f.getframerate()
            metadata.length = 1000 * f.getnframes() // f.getframerate()
            metadata['~format'] = self.NAME
            self._add_path_to_metadata(metadata)
            return metadata

        def _save(self, filename, metadata):
            log.debug("Saving file %r", filename)

        @classmethod
        def supports_tag(cls, name):
            return False



# -*- coding: utf-8 -*-
#
# Picard, the next-generation MusicBrainz tagger
#
# Copyright (C) 2004 Robert Kaye
# Copyright (C) 2007 Lukáš Lalinský
# Copyright (C) 2008 Will
# Copyright (C) 2008, 2018-2021 Philipp Wolfer
# Copyright (C) 2009 david
# Copyright (C) 2013 Johannes Dewender
# Copyright (C) 2013 Sebastian Ramacher
# Copyright (C) 2013, 2018-2021 Laurent Monin
# Copyright (C) 2013-2014 Michael Wiencek
# Copyright (C) 2016-2017 Sambhav Kothari
# Copyright (C) 2017 Sophist-UK
# Copyright (C) 2022 Bob Swift
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.


import os.path

from picard import log
from picard.config import get_config
from picard.const.sys import (
    IS_LINUX,
    IS_WIN,
)


try:
    from libdiscid.compat import discid
except ImportError:
    try:
        import discid
    except (ImportError, OSError):
        discid = None

DISCID_NOT_LOADED_MESSAGE = "CDROM: discid library not found - Lookup CD functionality disabled"

DEFAULT_DRIVES = []
if discid is not None:
    device = discid.get_default_device()
    if device:
        DEFAULT_DRIVES.append(device)

LINUX_CDROM_INFO = '/proc/sys/dev/cdrom/info'


def _generic_iter_drives():
    config = get_config()
    yield from (
        device.strip() for device
        in config.setting['cd_lookup_device'].split(',')
        if device and not device.isspace()
    )


def _parse_linux_cdrom_info(f):
    drive_names = []
    drive_audio_caps = []
    DRIVE_NAME = 'drive name:'
    CAN_PLAY_AUDIO = 'Can play audio:'
    for line in f:
        if line.startswith(DRIVE_NAME):
            drive_names = line[len(DRIVE_NAME):].split()
            break
    if drive_names:
        for line in f:
            if line.startswith(CAN_PLAY_AUDIO):
                drive_audio_caps = [v == '1' for v in line[len(CAN_PLAY_AUDIO):].split()]
                break
    yield from zip(drive_names, drive_audio_caps)


if IS_WIN:
    from ctypes import windll

    AUTO_DETECT_DRIVES = True
    DRIVE_TYPE_CDROM = 5

    def _iter_drives():
        GetLogicalDrives = windll.kernel32.GetLogicalDrives
        GetDriveType = windll.kernel32.GetDriveTypeW
        mask = GetLogicalDrives()
        for i in range(26):
            if mask >> i & 1:
                drive = chr(i + ord('A')) + ':'
                if GetDriveType(drive) == DRIVE_TYPE_CDROM:
                    yield drive

elif IS_LINUX and os.path.isfile(LINUX_CDROM_INFO):
    AUTO_DETECT_DRIVES = True

    def _iter_drives():
        # Read info from /proc/sys/dev/cdrom/info
        with open(LINUX_CDROM_INFO, 'r') as f:
            # Show only drives that are capable of playing audio
            yield from (
                os.path.realpath('/dev/%s' % drive)
                for drive, can_play_audio in _parse_linux_cdrom_info(f)
                if can_play_audio
            )

else:
    # There might be more drives we couldn't detect
    # setting uses a text field instead of a drop-down
    AUTO_DETECT_DRIVES = False
    _iter_drives = _generic_iter_drives


def get_cdrom_drives():
    """List available disc drives on the machine
    """
    # add default drive from libdiscid to the list
    drives = set(DEFAULT_DRIVES)
    try:
        drives |= set(_iter_drives())
    except OSError as error:
        log.error(error)
    return sorted(drives)
